/*
 * q_dsmark.c - Differentiated Services Marking qdisc
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <stddef.h>

#include "util.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "op.h"
#include "tc.h"
#include "filter.h"
#include "qdisc_common.h"
#include "qdisc.h"


#define __DEFAULT_PRM(f) f(mask)


/* ------------------------------- Checking -------------------------------- */


static void dsmark_check(QDISC *qdisc)
{
    DEFAULT_DECL;
    int tmp;
    CLASS *class;
    int got_qdisc = 0;

    param_get(qdisc->params);
    DEFAULT_SET;
    for (tmp = prm_indices.v; tmp != 1; tmp >>= 1)
	if (tmp & 1) error("indices must be a power of two");
    for (class = qdisc->classes; class; class = class->next) {
	check_childless(class);
	param_get(class->params);
	if (class->qdisc && (class->number != UNDEF_U32 ||
	  prm_mask.present || prm_value.present))
	    warn("suggest unnumbered and parameter-less class for dsmark's "
	      "inner qdisc");
	DEFAULT_GET;
	if (class->number == UNDEF_U32) {
	    if (class->qdisc) class->number = 0;
	    else error("dsmark does not auto-assign class numbers");
	}
	if ((prm_mask.present || prm_value.present) &&
	  class->number >= prm_indices.v)
	    errorf("dsmark class number (%lu) must be < indices (%lu) if "
	      "using parameters",(unsigned long) class->number,
	      (unsigned long) prm_indices.v-1);
	if (prm_mask.present && prm_mask.v > 0xff)
	    errorf("mask 0x%x above limit 0xff",prm_mask.v);
	if (!class->qdisc) continue;
	if (got_qdisc) error("dsmark has only one inner qdisc");
	check_qdisc(class->qdisc);
	got_qdisc = 1;
    }
    check_filters(qdisc->filters);
}


/* ------------------------ Default classification ------------------------- */


/*
 * @@@ Supporting dsmark's behaviour in the absence of a default index is
 * kind of difficult, because there are 64k possible skb->tc_index values
 * that need to be mapped to "classes". So it's better to wait with this
 * until we can absorb dsmark into some more abstract data forwarding
 * model.
 *
 * Update: at least if set_tc_index is present, things don't look quite so
 * bad.
 */

static void dsmark_default_class(DATA *d,QDISC *qdisc)
{
    DATA dfl;
    int i;

    param_get(qdisc->params);
    if (prm_default_index.present) {
	append_default(d,
	  data_decision(dr_class,require_class(qdisc,prm_default_index.v)));
	return;
    }
    if (!prm_set_tc_index.present || !prm_set_tc_index.v)
	yyerror("dsmark_default_class: need \"default_index\" or "
	  "\"set_tc_index\"");
    dfl = data_decision(dr_class,require_class(qdisc,255));
    for (i = 254; i >= 0; i--)
	dfl = op_binary(&op_logical_or,
	  op_binary(&op_logical_and,
	    op_binary(&op_eq,
	      op_ternary(&op_access,data_none(),data_unum(1),data_unum(8)),
	      data_unum(i)),
	    data_decision(dr_class,require_class(qdisc,i))),
          dfl);
    append_default(d,dfl);
}


/* -------------------------------- Dump tc -------------------------------- */


static void dsmark_dump_tc(QDISC *qdisc)
{
    DEFAULT_DECL;
    CLASS *class;

    tc_pragma(qdisc->params);
    param_get(qdisc->params);
    DEFAULT_SET;
    tc_qdisc_add(qdisc);
    tc_more(" indices %d",(int) prm_indices.v);
    if (prm_default_index.present)
	tc_add_unum("default_index",prm_default_index.v);
    if (prm_set_tc_index.present) tc_more(" set_tc_index");
    tc_nl();
    for (class = qdisc->classes; class; class = class->next) {
	tc_pragma(class->params);
	param_get(class->params);
	DEFAULT_GET;
	if (prm_mask.present || prm_value.present) {
	    tc_class_change(class);
	    if (prm_mask.present) tc_add_hex("mask",prm_mask.v);
	    if (prm_value.present) tc_add_hex("value",prm_value.v);
	    tc_nl();
	}
	dump_qdisc(class->qdisc);
    }
    dump_filters(qdisc->filters);
}


/* ------------------------------ Descriptors ------------------------------ */


static const PARAM_DSC *dsmark_qdisc_req[] = {
    &prm_indices,
    NULL
};

static const PARAM_DSC *dsmark_qdisc_opt[] = {
    &prm_default_index,
    &prm_pragma,	/* list */
    &prm_set_tc_index,
    DEFAULT_LIST	/* NB: no trailing comma */
    NULL
};

static const PARAM_DSC *dsmark_class_opt[] = {
    &prm_pragma,	/* list */
    &prm_value,
    DEFAULT_LIST	/* NB: no trailing comma */
    NULL
};

static PARAM_DEF dsmark_qdisc = {
    .required = dsmark_qdisc_req,
    .optional = dsmark_qdisc_opt,
};

static PARAM_DEF dsmark_class = {
    .required = NULL,
    .optional = dsmark_class_opt,
};

QDISC_DSC dsmark_dsc = {
    .name = "dsmark",
    .flags = QDISC_HAS_CLASSES | QDISC_HAS_FILTERS | QDISC_HAS_DEFAULT |
      QDISC_CHILD_QDISCS | QDISC_SHARED_QDISC,
    .qdisc_param = &dsmark_qdisc,
    .class_param = &dsmark_class,
    .check = &dsmark_check,
    .default_class = &dsmark_default_class,
    .dump_tc = &dsmark_dump_tc,
    .dump_ext = &generic_dump_ext,
};
